namespace StudySystemClient {
	public class Connection {
		private SessionManager session_manager;
		private TransactionManager transaction_manager;
		private Worker worker;

		public Connection(string cert_dir) throws Error {
			var loopback = new InetAddress.loopback(SocketFamily.IPV6);
			var session_factory
				= new SessionFactory(loopback, 12888, cert_dir);
			session_manager = new SessionManager(
				session_factory, (msg) => receive(msg));
			transaction_manager = new TransactionManager();
			worker = new Worker(session_manager);
		}

		public async Response.Body? send(Request.Body body) {
			var transaction_id = transaction_manager.register(send.callback);
			var request = new Request.Request(transaction_id, body);
			session_manager.send(request.encode());
			yield;
			return transaction_manager.get_result(transaction_id);
		}

		private void receive(owned uint8[] msg) {
			Response.Response response;
			try {
				response = new Response.Response.from_bytes(msg);
			} catch (Response.DecodeError e) {
				stderr.printf("Invalid response from server: %s", e.message);
				return;
			}
			Idle.add(() => {
					transaction_manager.resolve(response);
					return false;
				}, GLib.Priority.DEFAULT_IDLE);
		}
	}

	private class Continuation {
		private SourceFunc callback;

		public Continuation(owned SourceFunc callback) {
			this.callback = (owned) callback;
		}

		public void resume() {
			callback();
		}
	}

	private class TransactionManager {
		private uint16 next_transaction_id;
		private HashTable<uint16, Continuation> pending;
		private HashTable<uint16, Response.Body> results;

		public TransactionManager() {
			next_transaction_id = 0;
			pending = new HashTable<uint16, Continuation>(null, null);
			results = new HashTable<uint16, Response.Body>(null, null);
		}

		public uint16 register(owned SourceFunc callback) {
			var transaction_id = next_transaction_id++;
			pending.insert(transaction_id,
						   new Continuation((owned) callback));
			return transaction_id;
		}

		public void resolve(Response.Response response) {
			var transaction_id = (uint16)response.transaction_id;
			var continuation = pending.lookup(transaction_id);
			if (continuation == null) {
				stderr.printf("Response for non-pending transaction %d\n",
							  transaction_id);
				return;
			}
			results.insert(transaction_id, response.body);
			continuation.resume();
		}

		public Response.Body? get_result(uint16 transaction_id) {
			return results.lookup(transaction_id);
		}
	}

	private class Worker {
		private uint TASK_PERIOD_MS = 10;

		private SessionManager session_manager;
		private bool exit;
		private Thread<void> thread;

		public Worker(SessionManager session_manager) {
			this.session_manager = session_manager;
			exit = false;
			thread = new Thread<void>("connection_worker", body);
		}

		~Worker() {
			exit = true;
			thread.join();
		}

		private void body() {
			while (!exit) {
				session_manager.task();
				Thread.usleep(1000 * TASK_PERIOD_MS);
			}
		}
	}
}
